@@ -21,6 +21,7 @@ isFALSE <- function(x){
 #' @param diff   \code{[character]} difference between current and target value
 #'     (if any).
 #' @param short  \code{[character]} short description of the difference
+#' @param info   \code{[character]} other information, to be printed in the long message
 #' @param file   \code{[character]} File location of the test.
 #' @param fst    \code{[integer]} First line number in the test file.
 #' @param lst    \code{[integer]} Last line number in the test file (differs
@@ -54,17 +55,19 @@ isFALSE <- function(x){
 #' @export
 tinytest <- function(result, call
     , diff = NA_character_
-    , short= NA_character_
+    , short= c(NA_character_,"data","attr","xcpt", "envv","wdir","file")
+    , info = NA_character_
     , file = NA_character_
     , fst  = NA_integer_
     , lst  = NA_integer_
     ,...){
-
+  short <- match.arg(short)
   structure(result         # logical TRUE/FALSE
     , class    = "tinytest"
     , call     = call  # call creating or motivating the object
     , diff     = diff  # diff if isFALSE(result)
     , short    = short # short diff (4 char)
+    , info     = info  # user-defined info
     , file     = file  # test file location
     , fst      = fst   # first line of call
     , lst      = lst   # last line of call
@@ -75,8 +78,9 @@ tinytest <- function(result, call
 na_str <- function(x) if ( is.na(x) ) "" else as.character(x)
 
 oneline <- function(x) sub("\\n.+","...",x)
-indent <- function(x, with="     ")
+indent <- function(x, with="     "){
   gsub("\\n *",paste0("\n",with),paste0(with,sub("^ +","",x)))
+}
 
 lineformat <- function(x){
   if ( is.na(x) ) ""
@@ -105,19 +109,23 @@ format.tinytest <- function(x,type=c("long","short"), ...){
   file  <- na_str(d$file)
   short <- na_str(d$short)
   diff  <- d$diff
+  info  <- na_str(d$info)
 
   result <- if (isTRUE(x)) "PASSED      " 
             else if (isFALSE(x)) sprintf("FAILED[%s]",short)
             else if (is.na(x)  ) sprintf("SIDEFX[%s]",short)
   longfmt <- "----- %s: %s<%s--%s>\n%s"
+  # make room for diff and info fields when necessary
   if (isFALSE(x)||is.na(x)) longfmt <- paste0(longfmt, "\n%s")
+  if (!is.na(d$info)) longfmt <- paste0(longfmt,"\n%s")
 
   if (type == "short"){
     sprintf("%s: %s<%s--%s> %s", result, basename(file), fst, lst, oneline(call))
   }  else {
     sprintf(longfmt, result, file, fst, lst
                 , indent(call, with=" call| ")
-                , indent(diff, with=" diff| "))
+                , indent(diff, with=" diff| ")
+                , indent(info, with=" info| "))
   }
 
 }
@@ -166,8 +174,8 @@ longdiff <- function(current, target, alt){
        is_scalar(current) && 
        is_scalar(target) ){
        if ( all(class(current) %in% c("character","ordered","factor", "POSIXt","POSIXct")) ) 
-         sprintf("Expected '%s', got '%s'", current, target)
-       else sprintf("Expected %s, got %s", current, target)
+         sprintf("Expected '%s', got '%s'", target, current)
+       else sprintf("Expected %s, got %s", target, current)
   } else {
     paste0(" ", alt, collapse="\n")
   }
@@ -192,8 +200,11 @@ shortdiff <- function(current, target, ...){
 #'
 #' @param current \code{[R object or expression]} Outcome or expression under scrutiny.
 #' @param target \code{[R object or expression]} Expected outcome
-#' @param tol \code{[numeric]} Test equality to machine rounding. Passed
+#' @param tolerance \code{[numeric]} Test equality to machine rounding. Passed
 #'     to \code{\link[base]{all.equal} (tolerance)}
+#' @param info \code{[character]} scalar. Optional user-defined message. Must
+#'  be a single character string. Multiline comments may be separated by
+#'  \code{"\\n"}.
 #' @param ... Passed to \code{all.equal}
 #'
 #' @return A \code{\link{tinytest}} object. A tinytest object is a
@@ -222,27 +233,28 @@ shortdiff <- function(current, target, ...){
 #' expect_equal(2, c(x=2))      # FALSE
 #'
 #' @export
-expect_equal <- function(current, target, tol = sqrt(.Machine$double.eps), ...){
+expect_equal <- function(current, target, tolerance = sqrt(.Machine$double.eps), info=NA_character_, ...){
 
-  check <- all.equal(target, current, tol=tol, ...)
+  check <- all.equal(target, current, tolerance=tolerance, ...)
   equal <- isTRUE(check)
   diff  <- if (equal) NA_character_ else longdiff( current, target, check) 
-  short <- if (equal) NA_character_ else shortdiff(current, target, tolerance=tol)
+  short <- if (equal) NA_character_ else shortdiff(current, target, tolerance=tolerance)
 
-  tinytest(result = equal, call = sys.call(sys.parent(1)), diff=diff, short=short)
+  tinytest(result = equal, call = sys.call(sys.parent(1)), diff=diff, short=short, info=info)
 }
 
 
 
 #' @rdname expect_equal
 #' @export
-expect_identical <- function(current, target){
+expect_identical <- function(current, target, info=NA_character_){
   result <- identical(current, target)
   diff <-  if (result) NA_character_
            else longdiff(current, target, all.equal(target, current))
   short <- if (result) NA_character_
            else shortdiff(current, target, tolerance=0)
-  tinytest(result=result, call=sys.call(sys.parent(1)), diff=diff, short=short)
+  tinytest(result=result, call=sys.call(sys.parent(1)), diff=diff
+         , short=short, info=info)
 }
 
 
@@ -254,17 +266,18 @@ expect_identical <- function(current, target){
 #'
 #' @rdname expect_equal
 #' @export
-expect_equivalent <- function(current, target, tol = sqrt(.Machine$double.eps), ...){
+expect_equivalent <- function(current, target, tolerance = sqrt(.Machine$double.eps)
+                            , info=NA_character_, ...){
   out <- expect_equal(current, target
           , check.attributes=FALSE,use.names=FALSE
-          , tol=tol, ...)
+          , tolerance=tolerance, info=info, ...)
   attr(out, 'call') <- sys.call(sys.parent(1))
   out
 }
 
 #' @rdname expect_equal
 #' @export
-expect_true <- function(current){
+expect_true <- function(current, info=NA_character_){
   result <- isTRUE(current)
   call <- sys.call(sys.parent(1))
   if (!result){
@@ -273,15 +286,15 @@ expect_true <- function(current){
             else sprintf("object of class '%s'",class(current))
     diff  <- sprintf("Expected TRUE, got %s", this)
     short <- shortdiff(TRUE, FALSE)
-    tinytest(result, call=call,diff=diff, short=short)
+    tinytest(result, call=call,diff=diff, short=short, info=info)
   } else {
-    tinytest(result, call = sys.call(sys.parent(1)))
+    tinytest(result, call = sys.call(sys.parent(1)), info=info)
   }
 }
 
 #' @rdname expect_equal
 #' @export
-expect_false <- function(current){
+expect_false <- function(current, info=NA_character_){
   result <- isFALSE(current)
   call   <- sys.call(sys.parent(1))
   if (!result){
@@ -290,9 +303,9 @@ expect_false <- function(current){
             else sprintf("object of class '%s'",class(current))
     diff  <- "Expected FALSE, got TRUE"
     short <- shortdiff(TRUE, FALSE)
-    tinytest(result, call=call,diff=diff, short=short)
+    tinytest(result, call=call,diff=diff, short=short, info=info)
   } else {
-    tinytest(result, call = sys.call(sys.parent(1)))
+    tinytest(result, call = sys.call(sys.parent(1)), info=info)
   }
 }
 
@@ -316,7 +329,7 @@ expect_false <- function(current){
 #' expect_silent(print("hihi", quiet=FALSE)) # FALSE, and printed
 #'
 #' @export
-expect_silent <- function(current, quiet=TRUE){
+expect_silent <- function(current, quiet=TRUE, info=NA_character_){
 
   ## Make sure that printed output does not go to screen.
   # nullfile() was introduced at 3.6.0 and we want to be usable
@@ -366,19 +379,21 @@ expect_silent <- function(current, quiet=TRUE){
     , call  = sys.call(sys.parent(1))
     , short = if (result) NA_character_ else "xcpt"
     , diff  = diff
+    , info  = info
   )
 }
 
 
 #' @rdname expect_equal
 #' @export
-expect_null <- function(current){
+expect_null <- function(current, info=NA_character_){
   call <- sys.call(sys.parent(1))
   if (is.null(current)){
-    tinytest(TRUE, call=call)
+    tinytest(TRUE, call=call, info=info)
   } else {
     tinytest(FALSE, call=call, short="data"
       , diff = sprintf("Expected NULL, got '%s'", paste(class(current), collapse=", "))
+      , info = info
     )
   }
 }
@@ -386,36 +401,51 @@ expect_null <- function(current){
 
 #' @rdname expect_equal
 #' @param pattern \code{[character]} A regular expression to match the message.
+#' @param class \code{[character]} For condition signals (error, warning, message)
+#'        the class from which the condition should inherit.
 #' @export
-expect_error <- function(current, pattern=".*"){
+expect_error <- function(current, pattern=".*", class="error", info=NA_character_){
   result <- FALSE
   diff <- "No error"
   
   tryCatch(current, error=function(e){
-            if (grepl(pattern, e$message)){
-                result <<- TRUE
-            } else {
+            matches <- grepl(pattern, e$message)
+            isclass <- inherits(e, class)
+
+            if (matches && isclass){
+              result <<- TRUE
+            } else if (!isclass){
+              diff <<- sprintf("Error of class '%s', does not inherit from '%s'"
+                              , paste(class(e), collapse=", "), class)
+            } else if (!matches){
               diff <<- sprintf("The error message:\n '%s'\n does not match pattern '%s'"
-                             , e$message, pattern)
+                              , e$message, pattern)
             }
   })
   tinytest(result, call = sys.call(sys.parent(1))
            , short= if(result) NA_character_ else "xcpt"
-           , diff = if(result) NA_character_ else diff)
+           , diff = if(result) NA_character_ else diff
+           , info = info)
 }
 
 #' @rdname expect_equal
 #' @export
-expect_warning <- function(current, pattern=".*"){
+expect_warning <- function(current, pattern=".*", class="warning", info=NA_character_){
   
   result <- FALSE
   diff <- "No warning"
 
   withCallingHandlers(current
     , warning = function(w){
-        if (grepl(pattern, w$message)){
+        matches <- grepl(pattern, w$message)
+        isclass <- inherits(w, class)
+
+        if ( matches && isclass ){
           result <<- TRUE
-        } else {
+        } else if ( !isclass ){
+          diff <<- sprintf("Warning of class '%s', does not inherit from '%s'"
+                          , paste(class(w), collapse=", "), class)
+        } else if (!matches){
           diff <<- sprintf("The warning message\n '%s'\n does not match pattern '%s'"
                           , w$message, pattern)
         }
@@ -425,60 +455,183 @@ expect_warning <- function(current, pattern=".*"){
 
   tinytest(result, call=sys.call(sys.parent(1))
            , short = if (result) NA_character_ else "xcpt"
-           , diff  = if (result) NA_character_ else diff)
+           , diff  = if (result) NA_character_ else diff
+           , info  = info)
 }
 
 
 #' @rdname expect_equal
 #' @export
-expect_message <- function(current, pattern=".*"){
-  value <- ""
-  tc <- textConnection("value", open="w", local=TRUE)
-  sink(file=tc,type="message", split=FALSE)
+expect_message <- function(current, pattern=".*", class="message", info=NA_character_){
+#  value <- ""
+#  tc <- textConnection("value", open="w", local=TRUE)
+#  sink(file=tc,type="message", split=FALSE)
   
 
-  result <- TRUE
+  result <- FALSE
   msg    <- ""
   tryCatch(current
     , error = function(e){
-        result <<- FALSE
         msg <<- sprintf("Expected message, got error:\n '%s'",e$message)
       }
     , warning = function(w){
-        result <<- FALSE
         msg <<- paste(sprintf("Expected message, got warning:\n '%s'", w$message)
                     , collapse="\n")
       }
+    , message = function(m){
+        matches <- grepl(pattern, m$message)
+        isclass <- inherits(m, class)
+        if (matches && isclass){
+          result <<- TRUE
+        } else if (!isclass){
+          result <<- FALSE
+          msg <<- sprintf("Message of class '%s', does not inherit from '%s'"
+                          , paste(class(m), collapse=", "), class)
+        } else if (!matches){
+          msg <<- sprintf("The message message\n '%s'\n does not match pattern '%s'"
+                          , m$message, pattern)
+        }
+      }
   )
-  sink(file = NULL, type="message")
-  close(tc)
-  
-  # collapse the value string in case multiple messages were caught.
-  value <- paste(value, collapse="\n")
-  call <- sys.call(sys.parent(1))
-  # we got a warning or error instead of a message:
-  if (!result){
-    tinytest(
-        result
-      , call
-      , diff  = msg
-      , short = "xcpt" 
-    ) 
-  # we got a message, check if it matches 'pattern'
-  } else if (!isTRUE(grepl(pattern, value)) ){
-    df <- if (value == "") "No message"
-          else sprintf("The message\n '%s'\n doen not match pattern '%s'",value,pattern)
-    tinytest(FALSE
-      , call
-      , diff = df
-      , short = "xcpt"
-    )
-  } else {
-    tinytest(TRUE, call)
-  }
+#  sink(file = NULL, type="message")
+#  close(tc)
   
+  tinytest(result, call = sys.call(sys.parent(1))
+           , short= if(result) NA_character_ else "xcpt"
+           , diff = if(result) NA_character_ else msg
+           , info = info)
+#  # collapse the value string in case multiple messages were caught.
+#  value <- paste(value, collapse="\n")
+#  call <- sys.call(sys.parent(1))
+#  # we got a warning or error instead of a message:
+#  if (!result){
+#    tinytest(
+#        result
+#      , call
+#      , diff  = msg
+#      , short = "xcpt" 
+#      , info  = info
+#    ) 
+#  # we got a message, check if it matches 'pattern'
+#  } else if (!isTRUE(grepl(pattern, value)) ){
+#    df <- if (value == "") "No message"
+#          else sprintf("The message\n '%s'\n doen not match pattern '%s'",value,pattern)
+#    tinytest(FALSE
+#      , call
+#      , diff = df
+#      , short = "xcpt"
+#      , info  = info
+#    )
+#  } else {
+#    tinytest(TRUE, call, info=info)
+#  }
+#  
 }
 
+#' @rdname expect_equal
+#'
+#' @details
+#'
+#' \code{expect_stdout} Expects that output is written to \code{stdout},
+#' for example using \code{cat} or \code{print}. Use \code{pattern} to
+#' specify a regular expression matching the output.
+#'
+#'
+#' @export
+expect_stdout <- function(current, pattern=".*", info=NA_character_){
+  value <- ""
+  msg <- NA_character_
+  
+  tc <- textConnection("value", open="w", local=TRUE)
+  sink(file=tc, type="output", split=FALSE)
+    tryCatch(current
+      , error=function(e){sink(file=NULL, type="output"); stop(e)}
+    )
+  sink(file = NULL, type="output")
+  close(tc)
+
+  result <- grepl(pattern, paste0(value,""))
+  if (!result)
+    msg <- sprintf("output '%s'\n does not match pattern '%s'", value, pattern)
+
+  tinytest(result, call = sys.call(sys.parent(1))
+           , short= if(result) NA_character_ else "xcpt"
+           , diff = msg
+           , info = info)
+}
+
+
+#' Compare object with object stored in a file
+#'
+#' Compares the current value with a value stored to file with
+#' \code{\link{saveRDS}}.  If the  file does not exist, the current value is
+#' stored into file, and the test returns \code{expect_null(NULL)}.
+#'
+#' @param current \code{[R object or expression]} Outcome or expression under 
+#'        scrutiny.
+#' @param file \code{[character]} File where the \code{target} is stored. If 
+#'        \code{file} does not exist, \code{current} will be stored there.
+#' @param ... passed to \code{\link{expect_equal}}, respectively \code{\link{expect_equivalent}}.
+#'
+#' @note
+#' Be aware that on CRAN it is not allowed to write data to user space. So make
+#' sure that the file is either stored with your tests, or generated with
+#' \code{\link{tempfile}}, or the test is skipped on CRAN, using
+#' \code{\link{at_home}}.
+#' 
+#' Also note that \code{\link{build_install_test}} clones the package and
+#' builds and tests it in a separate R session in the background. This means
+#' that if you create a file located at \code{tempfile()} during the run, this
+#' file is destroyed when the separate R session is closed.
+#'
+#'
+#' @family test-functions
+#'
+#'
+#' @examples
+#' filename <- tempfile()
+#' # this gives TRUE: the file does not exist, but is created now.
+#' expect_equal_to_reference(1, file=filename)
+#' # this gives TRUE: the file now exists, and its contents is equal
+#' # to the current value
+#' expect_equal_to_reference(1, file=filename)
+#' # this gives FALSE: the file exists, but is contents is not equal
+#' # to the current value
+#' expect_equal_to_reference(2, file=filename)
+#'
+#' @export
+expect_equal_to_reference <- function(current, file, ...){
+  eetr(current=current, file=file, type="equal", ...)
+}
+
+#' @rdname expect_equal_to_reference
+#' @export
+expect_equivalent_to_reference <- function(current, file, ...){
+  eetr(current=current, file=file, type="equivalent", ...)
+}
+
+eetr <- function (current, file, type=c("equal","equivalent"), ...){
+
+    if (file.exists(file)){
+        out <- if (type=="equal")
+                  tinytest::expect_equal(current, readRDS(file), ...)
+                else
+                  tinytest::expect_equivalent(current, readRDS(file), ...)
+        if (!out){
+           diff <- attr(out, "diff")
+           diff <- paste(
+                    sprintf("current does not match target read from %s\n", file)
+                    , diff)
+           attr(out,"diff") <- diff
+        }
+        out
+    } else {
+        tinytest::expect_null(saveRDS(current, file)
+                , info=sprintf("Stored value in %s", file))
+    }
+}
+
+
 
 #' Report side effects for expressions in test files
 #'
@@ -487,6 +640,8 @@ expect_message <- function(current, pattern=".*"){
 #' @param report \code{[logical]} report all side-effects
 #' @param envvar \code{[logical]} changes in environment variables
 #' @param pwd    \code{[logical]} changes in working directory
+#' @param files  \code{[logical]} changes in files in the directory where the
+#'   test file lives. Also watches subdirectories.
 #'
 #' @section Details:
 #' A side effect causes a change in an external variable outside of the scope
@@ -517,12 +672,12 @@ expect_message <- function(current, pattern=".*"){
 #' report_side_effects(FALSE)
 #'
 #' # only report changes in environment variables
-#' report_side_effects(pwd=FALSE)
+#' report_side_effects(report=FALSE, envvar=TRUE)
 #'
 #' @export
-report_side_effects <- function(report=TRUE, envvar=report, pwd=report){
+report_side_effects <- function(report=TRUE, envvar=report, pwd=report, files=report){
   stopifnot(is.logical(envvar))
-  list(envvar=envvar, pwd=pwd)
+  list(envvar=envvar, pwd=pwd, files=files)
 } 
 
 # generate user-facing function that captures 'report_side_effects'
@@ -534,6 +689,10 @@ capture_se <- function(fun, env){
       env$envvar <- Sys.getenv()
     if (out[['pwd']])
       env$pwd <- getwd()
+    if (out[['files']]){
+      env$filesdir <- getwd()
+      env$files <- file.info(dir(env$filesdir, recursive=TRUE, full.names=TRUE))
+    }
     out
   }
 }
@@ -551,24 +710,6 @@ report_envvar <- function(env){
   out
 }
 
-# internal function, to be called by run_test_file after local capture.
-report_cwd <- function(env){
-  if ( !isTRUE(env$sidefx[['pwd']]) ) return(NULL)
-
-  old <- env$pwd
-  current <- getwd()
-  if ( identical(old, current) ) return(NULL)
-
-  msg <- sprintf("Working directory changed from \n '%s'\nto\n '%s'", old, current)
-  out <- tinytest(NA
-    , call = sys.call(sys.parent(1))
-    , short = "wdir"
-    , diff = msg
-  ) 
-  env$pwd <- current
-  out
-}
-
 
 
 
@@ -613,4 +754,64 @@ envdiff <- function(old, new){
 
 }
 
+# internal function, to be called by run_test_file after local capture.
+report_cwd <- function(env){
+  if ( !isTRUE(env$sidefx[['pwd']]) ) return(NULL)
+
+  old <- env$pwd
+  current <- getwd()
+  if ( identical(old, current) ) return(NULL)
+
+  msg <- sprintf("Working directory changed from \n '%s'\nto\n '%s'", old, current)
+  out <- tinytest(NA
+    , call = sys.call(sys.parent(1))
+    , short = "wdir"
+    , diff = msg
+  ) 
+  env$pwd <- current
+  out
+}
+
+
+
+report_files <- function(env){
+  if (!isTRUE(env$sidefx[['files']])) return(NULL)
+  old <- env$files
+  new <- file.info(dir(env$filesdir, recursive=TRUE, full.names=TRUE))
+
+  if ( identical(old, new) ) return(NULL)
+  on.exit(env$files <- new)
+
+  oldfiles <- rownames(old)
+  newfiles <- rownames(new)
+  
+  created <- setdiff(newfiles, oldfiles)
+  removed <- setdiff(oldfiles, newfiles)
+
+  remain  <- intersect(oldfiles, newfiles)
+  touched <- remain[old[remain,'mtime'] != new[remain, 'mtime']]
+
+  cre <- sprintf("Created: %s", if (length(created)>0) paste(created, collapse=", ") else character(0)) 
+  rem <- sprintf("Removed: %s", if (length(removed)>0) paste(removed, collapse=", ") else character(0))
+  alt <- sprintf("Touched: %s", if (length(touched)>0) paste(touched, collapse=", ") else character(0))
+  
+  diff <- paste(c(cre, rem, alt), collapse="\n")
+  # we do not record status changes, as they may mean different things
+  # on different OSs.
+  if (nchar(diff) == 0) return(NULL)
+  tinytest(NA
+    , call = sys.call(sys.parent(1))
+    , diff = diff
+    , short = "file"
+    , info  = "CRAN policy forbids writing in the package installation folder."
+  )
+
+}
+
+
+
+
+
+
+
 
